@@ -22,10 +22,17 @@
 -behaviour(gen_server).
 
 %% API
--export([new/1, new/2, start_link/2, delete/1, delete/2, delete/3,
-	 lookup/2, lookup/3, insert/3, insert/4, insert_new/3, insert_new/4,
-	 update/4, update/5, setopts/2, clear/1, clear/2, untag/1,
-	 clean/1, clean/2, filter/2, fold/3, all/0, info/1, info/2]).
+-export([start_link/2]).
+-export([new/1, new/2, delete/1]).
+-export([delete/2, delete/3, match_delete/2, match_delete/3]).
+-export([lookup/2, lookup/3]).
+-export([insert/3, insert/4, insert_new/3, insert_new/4]).
+-export([update/4, update/5]).
+-export([incr/3, incr/4]).
+-export([setopts/2, all/0, info/1, info/2]).
+-export([clear/1, clear/2, clean/1, clean/2]).
+-export([filter/2, fold/3]).
+-export([untag/1]).
 %% gen_server callbacks
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
 	 terminate/2, code_change/3]).
@@ -50,16 +57,19 @@
 -define(IS_EXPIRED(Time, LifeTime, CurrTime),
 	(is_integer(LifeTime) andalso (Time + LifeTime =< CurrTime))).
 
+-type tag() :: cache | nocache.
 -type counter() :: non_neg_integer().
 -type milli_seconds() :: pos_integer().
 -type option() :: {max_size, pos_integer() | infinity} |
 		  {life_time, milli_seconds() | infinity} |
-		  {cache_missed, boolean()}.
+		  {cache_missed, boolean()} |
+		  {type, set | ordered_set | bag | duplicate_bag}.
 -type read_fun() :: fun(() -> {ok, any()} | error).
 -type update_fun() :: fun(() -> ok | {ok, any()} | error).
 -type filter_fun() :: fun((any(), {ok, any()} | error) -> boolean()).
 -type fold_fun() :: fun((any(), {ok, any()} | error, any()) -> any()).
 -type state() :: #state{}.
+-export_type([tag/0]).
 
 %%%===================================================================
 %%% API
@@ -112,6 +122,19 @@ delete(Name, Key, Nodes) ->
 	      ets_delete(Name, Key)
       end, Nodes).
 
+-spec match_delete(atom(), any()) -> ok.
+match_delete(Name, Match) ->
+    match_delete(Name, Match, [node()]).
+
+-spec match_delete(atom(), any(), [atom()]) -> ok.
+match_delete(Name, Match, Nodes) ->
+    lists:foreach(
+      fun(Node) when Node /= node() ->
+	      send({Name, Node}, {match_delete, Match});
+	 (_MyNode) ->
+	      ets_match_delete(Name, Match)
+      end, Nodes).
+
 -spec insert(atom(), any(), any()) -> boolean().
 insert(Name, Key, Val) ->
     insert(Name, Key, Val, [node()]).
@@ -206,6 +229,20 @@ update(Name, Key, Val, UpdateFun, Nodes) ->
 	    Val
     end.
 
+-spec incr(atom(), any(), non_neg_integer()) -> ok.
+incr(Name, Key, Incr) ->
+    incr(Name, Key, Incr, [node()]).
+
+-spec incr(atom(), any(), non_neg_integer(), [node()]) -> ok.
+incr(Name, Key, Incr, Nodes) ->
+    Obj = {Key, Incr, current_time()},
+    lists:foreach(
+      fun(Node) when Node /= node() ->
+	      send({Name, Node}, {incr, Obj});
+	 (_) ->
+	      do_insert(incr, Name, Obj)
+      end, Nodes).
+
 -spec clear(atom()) -> ok.
 clear(Name) ->
     clear(Name, [node()]).
@@ -280,6 +317,7 @@ info(Name) ->
 		   ({owner, _}) -> true;
 		   ({name, _}) -> true;
 		   ({size, _}) -> true;
+		   ({type, _}) -> true;
 		   (_) -> false
 		end, ETSInfo)]
     end.
@@ -313,7 +351,8 @@ load_nif() ->
 %%%===================================================================
 init([Name, Opts]) ->
     process_flag(trap_exit, true),
-    ets:new(Name, [named_table, public, {read_concurrency, true}]),
+    Type = proplists:get_value(type, Opts, set),
+    ets:new(Name, [named_table, public, {read_concurrency, true}, Type]),
     Counter = new_counter(),
     TRef = start_clean_timer(),
     State = #state{tab = Name, clean_timer = TRef},
@@ -341,9 +380,15 @@ handle_info({insert, Op, {Key, _, CurrTime} = New}, #state{tab = Tab} = State) -
 	    do_insert(Op, Tab, New)
     end,
     {noreply, State};
+handle_info({incr, Obj}, #state{tab = Tab} = State) ->
+    do_insert(incr, Tab, Obj),
+    {noreply, State};
 handle_info({delete, Key}, #state{tab = Tab} = State) ->
     ets_delete(Tab, Key),
     {noreply, State};
+handle_info({match_delete, Match}, #state{tab = Tab} = State) ->
+    ets_match_delete(Tab, Match),
+    {noreply, State};
 handle_info(clear, #state{tab = Tab} = State) ->
     ets_delete_all_objects(Tab),
     {noreply, State};
@@ -423,7 +468,12 @@ do_insert(Op, Name, Obj, _CacheMissed, MaxSize) ->
     try
 	case Op of
 	    new -> ets:insert_new(Name, Obj);
-	    replace -> ets:insert(Name, Obj)
+	    replace -> ets:insert(Name, Obj);
+	    incr ->
+		incr_counter(Name),
+		{Key, Incr, Time} = Obj,
+		ets:update_counter(Name, Key, {2, Incr}),
+		ets:update_element(Name, Key, {3, Time})
 	end
     catch _:badarg ->
 	    false
@@ -485,8 +535,15 @@ check_opts(Opts) ->
 		  {life_time, Val} when is_integer(Val), Val>0 -> Opt;
 		  {life_time, infinity} -> Opt;
 		  {cache_missed, Bool} when is_boolean(Bool) -> Opt;
+		  {type, Type} when Type == set;
+				    Type == ordered_set;
+				    Type == bag;
+				    Type == duplicate_bag ->
+		      Opt;
 		  _ -> erlang:error({bad_option, Opt})
-	      end
+	      end;
+	 (cache_missed) -> {cache_missed, true};
+	 (Opt) -> erlang:error({bad_option, Opt})
       end, Opts).
 
 -spec do_setopts(state(), [{counter, counter()} | option()]) -> state().
@@ -513,7 +570,7 @@ do_setopts(State, Opts) ->
 
 -spec current_time() -> integer().
 current_time() ->
-    p1_time_compat:monotonic_time(milli_seconds).
+    erlang:system_time(millisecond).
 
 -spec start_clean_timer() -> reference().
 start_clean_timer() ->
@@ -572,6 +629,17 @@ ets_delete(Tab, Key) ->
 	    false
     end.
 
+-spec ets_match_delete(atom(), any()) -> boolean().
+ets_match_delete(Tab, Match) ->
+    try
+	incr_counter(Tab),
+	ets:match_delete(Tab, {Match, '_', '_'})
+    catch _:badarg ->
+	    false;
+	  _:{badmatch, undefined} ->
+	    false
+    end.
+
 -spec ets_delete_object(atom(), any()) -> boolean().
 ets_delete_object(Tab, Obj) ->
     try
